/*
 *	Copyright (c) 1994, 1995 by Jeremy Cooperstock and Steve Kotsopoulos,
 *	All Rights Reserved.
 */

/*
 * afdplib.c - AFDP utility functions
 *
 * static char rcsid[] = "$Id: afdplib.c,v 1.58 1998/07/02 19:37:36 steve Exp $";
 */

#include "sockutils.h"
#include "afdp.h"
#include "afdplib.h"

#define NUM_RETRIES 3

/*
 * Function to find a secretary
 * If lock is true, then assume calling process is prepared
 * to become secretary.
 * todo: should make this use multicast, with expanding-ring search
 */
int FindSecretary(group_id, secretary_host, admin_port, publish_port, lock, port)
group_id_t group_id;
host_t *secretary_host;
port_t *admin_port;
port_t *publish_port;
boolean_t lock;
port_t port;
{
	int sock;
	int on = 1;
	find_secretary_packet_t packet;
	int retry;
	int i;
	struct sockaddr_in sin;
	struct sockaddr_in from;
	int len;
	int status = FAIL;
	struct timeval timeout;
	fd_set r, w, e;
	struct hostent  *hp;

	logmsg(LOG_NOTICE,
	  "FindSecretary: searching for group %lu secretary [port %d] on %s.\n",
	  group_id, port, *secretary_host ?
	  HostaddrToName(*secretary_host) : "LAN" );

	/* build a find_secretary_packet to broadcast inquiry for secretary */
	packet.packet_type = FIND_SECRETARY_REQUEST;
	packet.group_id = htonl(group_id);
	packet.responsible = lock;

	/* multicast to the group of secretary the request */
	CreateSock(&sock, SOCK_DGRAM, SO_BROADCAST);
	if (setsockopt(sock,SOL_SOCKET,SO_REUSEADDR,(char *)&on,sizeof(on))< 0){
		fatal("FindSecretary: setsockopt: SO_REUSEADDR: %m");
	}
#ifdef SO_REUSEPORT
	if (setsockopt(sock,SOL_SOCKET,SO_REUSEPORT,(char *)&on,sizeof(on))< 0){
		fatal("FindSecretary: setsockopt: SO_REUSEPORT: %m");
	}
#endif

	/*
	 * If caller is prepared to become secretary, then first 
	 * make sure that no other host is about to become secretary.
	 * This is done by binding to the search port and listening
	 * for other potential FIND_SECRETARY_REQUEST packets.
	 * If another is found with higher host_id, then give it a 
	 * chance to startup.
	 */

	if (lock) 
	    BindSock (sock, &port, &sin);

	retry = 0;
	while (retry++ < NUM_RETRIES && status == FAIL) {
	    if(*secretary_host)
		UDPsend(*secretary_host, sock, port,
			(char *)&packet, sizeof(packet));
	    else
		Broadcast(sock, port, (char *)&packet, sizeof(packet));

	    /* might need to loop here to discard my own broadcast */
	    i = 0;
	    while (i++ < NUM_RETRIES && status == FAIL) {
		FD_ZERO(&r);
		FD_ZERO(&w);
		FD_ZERO(&e);
		FD_SET(sock, &r);
 		timeout.tv_sec =  1;
		timeout.tv_usec = 0;
		if (!select(getdtablesize(), &r, &w, &e, &timeout)) {
		    logmsg(LOG_DEBUG,"FindSecretary timeout...\n");
		}
		else if (FD_ISSET(sock, &r)) {
		    len = sizeof(from);
            	    recvfrom( sock, (char *) &packet, sizeof( packet ), 0,
				(struct sockaddr *) &from, &len);

        	    if (packet.packet_type == FIND_SECRETARY_REPLY) 
            		status = SUCCESS;
		/*
		 * Another secretary may be starting up if:
		 *
		 * we have just received a FIND_SECRETARY_REQUEST packet
		 * from a host willing to become a secretary and this host
		 * matches neither our IP address nor our hostname.
		 *
		 * In this situation, if the other secretary-candidate has 
	  	 * an IP address greater than ours, then let it become 
		 * secretary.
		 */
        	    else if (packet.packet_type == FIND_SECRETARY_REQUEST &&
		     packet.responsible == TRUE && 
		     strcmp(HostaddrToName(from.sin_addr.s_addr), Hostname()) &&
		     (from.sin_addr.s_addr) > Hostaddr()) {
			logmsg(LOG_INFO,"Found secretary elsewhere\n");
			Usleep(5);
            		retry = 0;
		    }
		}
		else {
		    fatal("impossible?\n");
		}
	    }
	} 

	if (status == SUCCESS) {
	    *admin_port = packet.admin_port;
	    *publish_port = packet.publish_port;

            hp = gethostbyaddr ((char *) &(from.sin_addr), 
		sizeof(from.sin_addr), AF_INET);
	    *secretary_host = from.sin_addr.s_addr;

            logmsg(LOG_NOTICE, "FindSecretary: got reply from %s (port %u)\n",
		(char *) HostaddrToName(from.sin_addr.s_addr),
		(port_t) ntohs(from.sin_port));
	    logmsg(LOG_DEBUG,"check: admin_port %u\n", ntohs(*admin_port));
	    logmsg(LOG_DEBUG,"check: publish_port %u\n", ntohs(*publish_port));
	}
	else {
	    logmsg(LOG_NOTICE, "Could not find secretary for group %lu\n", group_id);
	}
	close( sock );
	return( status );
}

#define CONTROL_TIMEOUT 5    /* number of seconds for timeout on reading port */

int TCPexchange(host, sock, port, packet, packet_size )
host_t host;
int    *sock;
port_t port;
char   *packet;
int    packet_size;
{
	struct sockaddr_in sin;
	int st;
	fd_set readfds, writefds, exceptfds;
	int nfds;
	struct timeval timeout;

	CreateSock( sock, SOCK_STREAM, 0 );
	sin.sin_family = AF_INET;
	sin.sin_port = port;
	sin.sin_addr.s_addr = host;

	if (connect(*sock, (struct sockaddr *) &sin, sizeof(sin)) < 0) {
		logmsg(LOG_ERR, "TCPexchange: can't connect to host %s:%u\n",
				HostaddrToName(host), (u_short) ntohs(port));
		return( FAIL );
	}
#ifdef not_IDENT
        logmsg(LOG_INFO, "IDENT: TCPexchange: secretary is %s@%s\n",
                PeerUser(tempsock), HostaddrToName(PeerHost(tempsock)));
#endif

	st = WriteStream(*sock, packet, packet_size);
	if (st != packet_size ) {
		logmsg(LOG_ERR, "TCPexchange can't write to port\n");
		return( FAIL );
	}

	nfds = getdtablesize();
        FD_ZERO( &readfds );
	FD_ZERO( &writefds );
	FD_ZERO( &exceptfds );
	FD_SET( *sock, &readfds );

	timeout.tv_sec = CONTROL_TIMEOUT;
	timeout.tv_usec = 0;
	st = select( nfds, &readfds, &writefds, &exceptfds, &timeout );
	if (!st) {
		logmsg(LOG_ERR,"TCPexchange: no reply\n" );
		return( FAIL );
	}

	st = ReadStream( *sock, packet, packet_size);
	if (st != packet_size) {
		logmsg(LOG_ERR,"TCPexchange can't read reply\n");
		return( FAIL );
	}
	return( SUCCESS );
}


int TCPsend(host, port, packet, packet_size)
host_t host;
port_t port;
char *packet;
int packet_size;
{
	int sock;
	struct sockaddr_in sin;
	int st;

	CreateSock( &sock, SOCK_STREAM, 0 );
	sin.sin_family = AF_INET;
	sin.sin_port = port;
	sin.sin_addr.s_addr = host;

	if (connect(sock, (struct sockaddr *) &sin, sizeof(sin)) < 0) {
		logmsg(LOG_ERR, "TCPsend: can't connect to host %s:%u\n",
				HostaddrToName(host), (u_short) ntohs(port));
		close (sock);
		return(-1);
	}
	st = WriteStream(sock, packet, packet_size);
	if (st != packet_size) {
		logmsg(LOG_ERR, "TCPsend: wrote wrong number of bytes\n");
		return(st);
	}
	close( sock );
	return(st);
}

host_t MulticastAddr(grpid)
group_id_t grpid;
{
	static char ip[20];
	sprintf(ip, "225.0.0.%d", (int)(grpid % 255) );
	return(HostnameToAddr(ip));
}
