/*
 *	Copyright (c) 1994, 1995 by Jeremy Cooperstock and Steve Kotsopoulos,
 *	All Rights Reserved.
 */

/*
 * secretary.c - AFDP secretary functions
 *
 * static char rcsid[] = "$Id: secretary.c,v 1.43 1998/07/02 19:37:36 steve Exp $";
 */

#include "config.h"
#include "sockutils.h"
#include "afdp.h"
#include "afdplib.h"
#include "afdpjoin.h"

/* secretary globals */
msg_id_t msg_id = 0; 			/* secretary supplies and increments */
u_int threshold = DEF_THRESHOLD;	/* when to switch mode */
u_char multicast_ttl = 1;		/* for multicasting over Internet   */

/* definition of our table of hosts in our group */
u_int maxmember = 0;            /* next free location in members table    */
u_int members   = 0;            /* number of members currently in group   */

struct {
	host_t addr;		/* subscriber addr in network-byte order  */
	char user[MAX_NAMELEN];	/* login name of user			  */
	u_char capability;	/* what multicasting abilities available  */
	u_char hops;		/* used for setting ttl for multicasts    */ 
} memberinfo[MAXMEMBERS];	/* entry 0 is always empty 		  */

/* definition of our table of active clients */
u_int maxclient = 0;		/* next free location in client table	  */
u_int clients   = 0;		/* number of clients currently xmitting   */

struct {
	host_t addr;		/* client addr in network-byte order 	  */
	u_long msize;		/* size of message being transmitted 	  */
} clientinfo[MAXCLIENTS];	/* entry 0 is always empty                */

/* interval timing globals */
time_t broadcast_tpp = DEF_TPP;
time_t multicast_tpp = DEF_TPP;
time_t unicast_tpp = DEF_TPP;

int IsMember(), AddMember(), DelMember(), IsClient(), AddClient(), DelClient();
void UpdateInterval();
time_t CalcInterval();

void HandleSearchPacket()
{
	int len, got, sent;
	find_secretary_packet_t packet;
	struct sockaddr_in from;

	len = sizeof (from);
	got = recvfrom(search_sock, (char *)&packet, 
		sizeof (find_secretary_packet_t), 0,
			(struct sockaddr *)&from, &len);
	if (got != sizeof(find_secretary_packet_t)) {
		logmsg(LOG_ERR,"HandleSearchPacket: only got %d bytes\n", got);
		return;
	}

	logmsg(LOG_INFO,"HandleSearchPacket: got packet from %s (port %u)\n",
		HostaddrToName(from.sin_addr.s_addr), ntohs(from.sin_port));

	if (packet.packet_type != FIND_SECRETARY_REQUEST) {
		logmsg(LOG_ERR,"packet type not FIND_SECRETARY_REQUEST... ignoring\n");
		logmsg(LOG_ERR,"packet type is %d\n", packet.packet_type);
		return;
	}

	if (ntohl (packet.group_id) != mygroup) {
		return; /* ignore (silently) packets for other groups */
	}

	packet.packet_type = FIND_SECRETARY_REPLY;
	packet.admin_port = admin_port;
	packet.publish_port = publish_port;

       	sent = sendto(search_sock, (char *)&packet, 
	   sizeof(find_secretary_packet_t), 0, (struct sockaddr *)&from, len); 
	if (sent != sizeof (find_secretary_packet_t)) 
		logmsg(LOG_NOTICE,"error responding to find_secretary_packet");
}

void HandleAdminPacket(sock)
int sock;
{
	int got, sent, i;
	host_t host;		/* note: we keep this in network-byte order */
	admin_packet_t packet;
	struct sockaddr_in peer;

	GetPeer(sock, &peer);
	logmsg(LOG_INFO,"HandleAdminPacket: got message from %s (port %u)\n",
		HostaddrToName(peer.sin_addr.s_addr), ntohs(peer.sin_port));

	host = peer.sin_addr.s_addr;	/* note: no conversion done */

	got = ReadStream (sock, (char *)&packet, sizeof(admin_packet_t));
	if (got != sizeof (admin_packet_t) ) {
		logmsg(LOG_ERR,"HandleAdminPacket: only got %d bytes\n", got);
		return;
	}

	if (ntohl(packet.group_id) != mygroup) {
		return; /* ignore (silently) packets for other groups */
	}

	switch(packet.packet_type) {

	case JOIN_REQUEST:
		if (IsMember(host)) {
			/* host already in this group */
			/* return SUCCESS anyways, in case sub rebooted */
			logmsg(LOG_NOTICE,"host %s is already in this group\n", 
				  HostaddrToName(host));
		}
		else {
			/* add host to memberinfo table */
			AddMember(host, packet.capability);
		}

		packet.packet_type = JOIN_REPLY;
		packet.group_data_port = group_data_port;
		packet.time = htonl(time(0));
		packet.status = SUCCESS;
		break;

	case LEAVE_REQUEST:
		packet.packet_type = LEAVE_REPLY;
		if ((i = IsMember(host))) {
			DelMember(i);
			packet.status = SUCCESS;
		}
		else {
			logmsg(LOG_NOTICE,"host %s is not in this group\n", 
				HostaddrToName(host));
			packet.status = FAIL;
		}
		break;

	default:
		/* detect impossible condition */
		/* could be caused by error in byte-swapping? */
		fatal("bad packet_type");
		break;
	}

       	sent = WriteStream(sock, (char *)&packet, sizeof(admin_packet_t));
	if (sent != sizeof(admin_packet_t)) 
		fatal("responding to admin_packet");
}

void HandlePublishPacket(sock)
int sock;
{
	int got, sent, i;
	u_short list_size = 0;
	host_t list[MAXMEMBERS];
	host_t host;
	publish_packet_t packet;
	struct sockaddr_in peer;
	u_long msize;

	GetPeer(sock, &peer);
	host = peer.sin_addr.s_addr;	/* note: no conversion done */

	logmsg(LOG_INFO,"HandlePublishPacket: got message from %s (port %u)\n",
		HostaddrToName(peer.sin_addr.s_addr), ntohs(peer.sin_port));

	got = ReadStream (sock, (char *)&packet, sizeof(publish_packet_t));
	if (got != sizeof (publish_packet_t) ) {
		logmsg(LOG_ERR,"HandlePublishPacket: only got %d bytes\n", got);
		return;
	}

	if (ntohl(packet.group_id) != mygroup) {
		return; /* ignore (silently) packets for other groups */
	}

	switch(packet.packet_type) {

	case PUBLISH_REQUEST:
		logmsg(LOG_NOTICE,"Received a publish request from %s for %lu bytes\n",
			HostaddrToName(peer.sin_addr.s_addr), 
			ntohl(packet.msg_size));
		msize = ntohl(packet.msg_size);

		/* now build reply */
		packet.packet_type = PUBLISH_PERMISSION;

		if (clients >= MAXCLIENTS - 2) {
			/* max clients already active */
			logmsg(LOG_NOTICE,"host %s told to wait\n",
				  HostaddrToName(host));
			packet.status = BUSY;
			packet.time = htonl(CLIENT_WAIT);
		}
		else if (publisher_host && (host != publisher_host)) {
			logmsg(LOG_NOTICE,"publisher host %s rejected\n",
				  HostaddrToName(host));
			packet.status = REJECT;
		}
		else {
			if (IsClient(host)) {
				/* allow it anyway */
				logmsg(LOG_NOTICE,
				"host %s is already an active client\n",
				HostaddrToName(host));
			}
			logmsg(LOG_NOTICE,"host %s given permission\n",
				  HostaddrToName(host));
			packet.status = OK;
			packet.time = htonl(time(0));

			packet.msg_id = htons(msg_id++);
			packet.group_data_port = group_data_port;

			packet.publish_mode = CalcMode(packet.publish_mode, 
							members);
			packet.interval = 
				htonl(CalcInterval(packet.publish_mode));

			packet.multicast_address = MulticastAddr(mygroup);
			packet.multicast_ttl = multicast_ttl;

			if (packet.publish_mode != MODE_BROADCAST) 
				list_size = 
					CalcUniList(list, packet.publish_mode);
			else 
				list_size = 0;

			packet.host_list_size = htons(list_size);
		}

		/* first send reply */
		sent = WriteStream(sock, (char *)&packet, 
			sizeof(publish_packet_t));
		if (sent != sizeof(publish_packet_t)) 
			fatal("responding to publish_packet");

		/*
		 * if we granted permission, add host to clientinfo and 
		 * check if we also need to send member list for unicasting 
		 */

		if (packet.status == OK)  {
			AddClient(host, msize);

			if (packet.publish_mode != MODE_BROADCAST && 
			    list_size != 0) {
				logmsg(LOG_INFO,"Send list of %u hosts to client\n",
					ntohs(packet.host_list_size));
				logmsg(LOG_INFO,"1st %s\n",
					HostaddrToName(list[0]));

				/* now send subscriber list */
				sent = WriteStream(sock, (char *)list, 
					list_size * sizeof(host_t)); 
				if (sent != list_size * sizeof(host_t)) 
					fatal("sending subscriber list");
			}
		}
		break;

	case PUBLISH_COMPLETE:
		logmsg(LOG_NOTICE,"Message %u completed by %s!\n",
			ntohs(packet.msg_id),
			HostaddrToName(peer.sin_addr.s_addr));

		if ((i = IsClient(host))) {
			UpdateInterval(packet.publish_mode, ntohl(packet.time), 
				clientinfo[i].msize);
			DelClient(i);
		}
		else 
			logmsg(LOG_NOTICE,"cannot find host %s in client list!\n",
				HostaddrToName(host));
		break; 

	default:
		/* detect impossible condition */
		fatal("bad packet_type");
		break;
	}
}

/*
 * determine appropriate publishing mode
 * if the publisher and all subscribers support multicast, use it
 * otherwise, use broadcast if the number of subscribers > threshold
 * otherwise, use unicast
 */
u_char CalcMode(mode, members)
u_char mode;
int members;
{
	u_int i;

	if (mode == MODE_MULTICAST) {
		for (i = 1 ; i <= maxmember ; i++) {
			if (memberinfo[i].addr != NULLHOST && 
			    memberinfo[i].capability != MODE_MULTICAST) {
				mode = MODE_BROADCAST;
				break;
			}
		}
	}

	if (mode == MODE_BROADCAST && members < threshold)
		mode = MODE_UNICAST;

	return mode;
}

/*
 * calculate appropriate client interval
 */
time_t CalcInterval(mode)
u_char mode;
{
	
	switch(mode) {

	case MODE_BROADCAST:
		return (broadcast_tpp * 0.5);
		break;
	case MODE_MULTICAST:
		return (multicast_tpp * 0.5);
		break;
	case MODE_UNICAST:
		return (unicast_tpp * 0.5);
		break;
	default:
		fatal("no such mode");
		return(0);
		break;
	}
}

void UpdateInterval(mode, time, msize)
u_char mode;
time_t time;
u_long msize;
{
	time_t tpp = 0;

	switch(mode) {

	case MODE_BROADCAST:
		tpp = broadcast_tpp = (time/msize) * MAX_BROADCAST_DATA_SIZE;
		break;
	case MODE_MULTICAST:
		tpp = multicast_tpp = (time/msize) * MAX_MULTICAST_DATA_SIZE;
		break;
	case MODE_UNICAST:
		tpp = unicast_tpp = (time/msize) * MAX_UNICAST_DATA_SIZE;
		break;
	default:
		fatal("no such mode");
		break;
	}
	logmsg(LOG_NOTICE,"Client sent %ld bytes in %ld ms, tpp = %ld ms\n",
		msize, time/mSEC, tpp/mSEC);
}

/*
 * calculate appropriate subscriber unicast list
 */
u_short CalcUniList(list, mode)
host_t list[];
u_char mode;
{
	int i;
	int count = 0;

	for (i = 1; i <= maxmember; i++)
		if ( memberinfo[i].addr != NULLHOST &&
		     (mode == MODE_UNICAST || 
		      memberinfo[i].capability != MODE_MULTICAST) ) 
			memcpy (&(list[count++]),
				(char *) &(memberinfo[i].addr),
				sizeof(host_t));
	return (count);
}

/*
 * utility functions for dealing with subscriber info and list
 * Note that 'host' is in network byte order.
 */
int IsMember(host)
host_t host;
{
	u_int i;

	for (i = 1 ; i <= maxmember ; i++) {
		if (memberinfo[i].addr == host) {
			return(i); /* found a match */
		}
	}
	return(0); /* did not find a match */
}

int AddMember(host, capability)
host_t host;
u_char capability;
{
	u_int i;

	for (i = 1 ; i <= maxmember+1 ; i++) {
		if (memberinfo[i].addr == NULLHOST)
			break;
	}
	if (i > maxmember)
		maxmember = i;

	if (maxmember >= (MAXMEMBERS - 1)) {
		fatal("members table full, increase MAXMEMBERS & recompile\n");
	}

	/* add to memberinfo */
	memberinfo[i].addr = host;
	memberinfo[i].capability = capability;

	members++;
	logmsg(LOG_NOTICE,"added host %s to position %u in group list\n", 
		HostaddrToName(host), i);
	return(i); /* return the position */
}

int DelMember(i)
u_int i;
{
	logmsg(LOG_NOTICE,"deleting host %s from position %u in group list\n", 
		HostaddrToName(memberinfo[i].addr), i);
	memberinfo[i].addr = NULLHOST;
	members--;
	return(1);
}

/*
 * utility functions for dealing with client info
 */
int IsClient(host)
host_t host;
{
	u_int i;

	for (i = 1 ; i <= maxclient ; i++) {
		if (clientinfo[i].addr == host) {
			return(i); /* found a match */
		}
	}
	return(0); /* did not find a match */
}

int AddClient(host, msize)
host_t host;
u_long msize;
{
	u_int i;

	for (i = 1 ; i <= maxclient+1 ; i++) {
		if (clientinfo[i].addr == NULLHOST)
			break;
	}
	if (i > maxclient)
		maxclient = i;

	if (maxclient >= (MAXCLIENTS - 1)) {
		fatal("client table is full, increase MAXCLIENTS\n");
	}
	clientinfo[i].addr = host;
	clientinfo[i].msize = msize;
	clients++;
	logmsg(LOG_NOTICE,"added client %s to position %u in client list\n", 
		HostaddrToName(host), i);
	return(i); /* return the position */
}

int DelClient(i)
u_int i;
{
	/* might want to add garbage collection for this later */
	logmsg(LOG_NOTICE,"deleting host %s from position %u in client list\n", 
		HostaddrToName(clientinfo[i].addr), i);
	clientinfo[i].addr = NULLHOST;
	clients--;
	return(1);
}
