/*
 *	Copyright (c) 1994, 1995 by Jeremy Cooperstock and Steve Kotsopoulos,
 *	All Rights Reserved.
 */

/*
 * sockutils.c - a collection of socket convenience functions
 *
 * static char rcsid[] = "$Id: sockutils.c,v 1.64 1998/07/02 19:37:36 steve Exp $";
 */

#include "sockutils.h"
#include "authuser.h"
#include <ctype.h>
#ifndef MIPS
#include <unistd.h>
#endif

/*
 * Create a socket (either UDP or TCP - specified by type).
 * If this socket will be broadcasting, set option to SO_BROADCAST.
 */ 
void CreateSock (sock, type, option)
int *sock;
int type;
int option;
{
	int on = 1;

	if ((*sock = socket(AF_INET, type, 0)) < 0) {
		fatal("CreateSock: socket: %m");
	}

	if (option)
	    if (setsockopt(*sock,SOL_SOCKET,option, (char *)&on,sizeof(on))< 0){
		fatal("CreateSock: setsockopt: %m");
	    }
}

/*
 * Bind a socket to a port.
 * If the port is unspecified (0) a port number will be assigned.
 */
void BindSock (sock, port, sin)
int sock;
u_short *port;			/* NOTE: port must be in network byte-order */
struct sockaddr_in *sin;
{
	int retry = 0;

	if ( (*port) == 0 ) {
		*port = AssignPort();	/* assign an arbitrary port number */
		retry = 1;	/* retry bind with new port, if it fails   */
	}

	bzero((char *)sin, sizeof(struct sockaddr_in));
	sin->sin_family      = AF_INET;
	sin->sin_addr.s_addr = htonl(INADDR_ANY);
	sin->sin_port        = *port;

	while (bind(sock,(struct sockaddr*)sin,sizeof(struct sockaddr_in))< 0) {
		if (retry) {
			logmsg(LOG_DEBUG,"BindSock: will call AssignPort again\n");
			sin->sin_port = *port = AssignPort();
		}
		else {
			fatal("BindSock: bind to port %d failed: %m", ntohs(*port));
		}
	}
	logmsg(LOG_INFO,"BindSock: port=%u\n", (u_short) ntohs(sin->sin_port));
}

/*
 * For TCP sockets, activate listening.
 */
void ListenSock(sock)
int sock;
{
	if (listen (sock, 5) < 0) {
		fatal("ListenSock: listen: %m");
	}
}

/*
 * Assign a new port number, return it in network byte-order.
 */
u_short AssignPort()
{
	static u_short count = 9999;

	logmsg(LOG_DEBUG,"assigning port %u\n", count);
	return htons(count--);
}

void Fatal(msg)		/* print fatal error message and die */
char *msg;
{
	perror(msg);
	exit(6);
}

/*
 * lookup service in /etc/services and return the port,
 *   if that fails check if service was a number and convert it
 *   if that fails return fallback_port
 */
u_short GetPort(service, protocol, fallback_port)
char *service, *protocol;
u_short fallback_port;
{
	struct servent *pse;
	u_short port;

	if ((pse = getservbyname(service, protocol)))
		return (port = pse->s_port);
	else {
		if ((port = htons((u_short)atoi(service))))
			return port;
		else if (fallback_port)
			return fallback_port;
		else
			fatal("GetPort: %m");
			return(0);		/* NOTREACHED */
	}
}

void GetPeer(sock, peer)  /* finds the remote peer of a connected socket */
int sock;
struct sockaddr_in *peer;
{
	int len;

	len = sizeof(struct sockaddr_in);
	if (getpeername(sock, (struct sockaddr *) peer, &len) < 0)
		fatal("GetPeer: getpeername: %m");
}

u_long PeerHost(sock)	/* returns the remote host in network byte-order */
int sock;
{
	struct sockaddr_in peer;

	GetPeer(sock, &peer);
	return(peer.sin_addr.s_addr);
}

u_short PeerPort(sock)	/* returns the remote port in network byte-order */
int sock;
{
	struct sockaddr_in peer;

	GetPeer(sock, &peer);
	return(peer.sin_port);
}

char * PeerUser(sock)	/* returns a pointer to the remote user's login name */
int sock;
{
	u_long inlocal, inremote;
	u_short local, remote;
	char *p, tmp[64];

	if (auth_fd2(sock, &inlocal, &inremote, &local, &remote) == -1)
		logmsg(LOG_ERR,"PeerUser: auth_fd2");
	strcpy(tmp, HostaddrToName(inlocal));
	logmsg(LOG_NOTICE,"IDENT: local is %d@%s, remote is %d@%s\n",
		ntohs(local), tmp,
		ntohs(remote), HostaddrToName(inremote));
	/* lookup the remote user name, timeout after 10 seconds */
	p = auth_tcpuser3(inlocal, inremote, local, remote, 10);
	if (p == NULL) {
		logmsg(LOG_ERR,"PeerUser: auth_tcpuser3");
		return("unknown");
	}
	return(p);
}

int UDPsend(host, sock, port, msg, msg_size)
u_long host;			/* NOTE: host must be in network byte-order */
int sock;
u_short port;			/* NOTE: port must be in network byte-order */
char *msg;
long msg_size;
{
	int st = 1;
	int sent;
	int len;
	struct sockaddr_in sin;

	sin.sin_family = AF_INET;
	sin.sin_port = port;
	sin.sin_addr.s_addr = host;

	len = sizeof( struct sockaddr_in);
	sent = sendto(sock, msg, msg_size, 0, (struct sockaddr *)&sin, len);
	logmsg(LOG_INFO,"UDPsend: sent %d bytes to %s (port %u)\n", sent,
		HostaddrToName(sin.sin_addr.s_addr), ntohs(sin.sin_port));

	if (sent != msg_size) {
		st = 0;
		logmsg(LOG_ERR,"UDPsend: wrong number of bytes sent: %d\n",sent);
	}
	return( st );
}

/*
 * We communicate with each neighbor in
 * a list constructed at the time we're
 * started up.  Neighbors are currently
 * directly connected via a hardware interface.
 */
struct	neighbor {
	struct	neighbor *n_next;
	char	*n_name;		/* interface name */
	char	*n_addr;		/* who to send to */
	int	n_addrlen;		/* size of address */
	int	n_flags;		/* should forward?, interface flags */
};

struct	neighbor *neighbors;

int Broadcast(sock, port, msg, msg_size)
int sock;
u_short port;			/* NOTE: port must be in network byte-order */
char *msg;
long msg_size;
{
	static int configured = FALSE;
	int st = 1;
	int sent;
	struct neighbor *np;
	struct sockaddr_in *sin;

	if (!configured) {
		/* ConfigureBroadcast() must be called once, */
		/* before any broadcast packets can be sent. */
		if (ConfigureBroadcast(sock) != 1)
			fatal("Broadcast: ConfigureBroadcast");

		configured = TRUE;
	}

	for (np=neighbors; np != NULL; np=np->n_next) {
		sin = (struct sockaddr_in *)np->n_addr;
		sin->sin_port = port;

		sent = sendto(sock, msg, msg_size, 0,
				(struct sockaddr *)np->n_addr, np->n_addrlen);
		logmsg(LOG_INFO,"Broadcast: sent %d bytes to LAN %s (port %u)\n",
				sent, HostaddrToName(sin->sin_addr.s_addr),
				ntohs(sin->sin_port));

		if (sent != msg_size) {
			st = 0;
			logmsg(LOG_ERR,"Broadcast should have sent: %ld: %m", msg_size);
			break;
		}
	} 
	return( st );
}



char * Hostname()	/* return's the hostname we are running on */
{
	static char	myname[MAXHOSTNAMELEN];

	if (gethostname(myname, sizeof (myname) - 1) < 0) {
		fatal("Hostname: gethostname: %m");
	}
	return myname;
}

u_long Hostaddr()	/* return's the host IP address, in network order */
{
	static char	myname[MAXHOSTNAMELEN];
	struct hostent	*myhp;

	strcpy(myname, Hostname());
	myhp = gethostbyname(myname);
	if (myhp == NULL) {
		logmsg(LOG_ERR,"couldn't lookup %s\n", myname);
		fatal("Hostaddr: gethostbyname: %m");
	}
	return ( *(u_long *) myhp->h_addr);
}

char * HostaddrToName(addr)	/* convert host address to hostname      */
u_long addr;			/* address must be in network byte-order */
{
	struct sockaddr_in from;
	struct hostent  *hp;
	struct in_addr tempaddr;

	from.sin_addr.s_addr = addr;
	hp = gethostbyaddr ((char *)&(from.sin_addr),
				sizeof(from.sin_addr), AF_INET);
	if (hp == NULL) {
		if (addr == 0)	/* why would you ask for that address? */
			return("NULLHOST");
		else { /* return a dotted-quad string if we can't find name */
			tempaddr.s_addr = addr;
			return(inet_ntoa(tempaddr));
		}
	}
	return(hp->h_name);
}

u_long HostnameToAddr(name)	/* returns address in network byte-order */
char *name;
{
	u_long tmp;

	if (isalpha(name[0])) {			/* input is a hostname */
		struct hostent	*hp;

		hp = gethostbyname(name);
		if (hp == NULL) {
			fatal("HostnameToAddr: gethostbyname: %m");
		}
		return ( *(u_long *) hp->h_addr);
	}
	else	/* input is a dotted-quad IP number, eg. "128.100.8.6" */
		if ((tmp = inet_addr(name)) == INADDR_NONE)
			fatal("HostnameToAddr: inet_addr: %m");
		return(tmp);
}

int ReadStream(fd, ptr, nbytes) /* from page 279 of UNIX Network Programming */
int fd, nbytes;
char *ptr;
{
	int nleft, nread;

	/* logmsg(LOG_DEBUG,"reading %d bytes\n", nbytes); */

	nleft = nbytes;
	while (nleft > 0) {
		nread = read(fd, ptr, nleft);
		/* logmsg(LOG_DEBUG,"read got %d bytes\n", nread); */
		if (nread < 0) {
			logmsg(LOG_ERR,"Readstream: read: %m");
			return(nread);		/* error, return < 0 */
		}
		else if (nread == 0)
			break;			/* EOF */

		nleft -= nread;
		ptr += nread;
	}
	return(nbytes - nleft);
}

int WriteStream(fd, ptr, nbytes)
int fd, nbytes;
char *ptr;
{
	int nleft, nwritten;

	nleft = nbytes;
	while (nleft > 0) {
		nwritten = write(fd, ptr, nleft);
		/* logmsg(LOG_DEBUG,"write wrote %d bytes\n", nwritten); */
		if (nwritten <= 0) {
			logmsg(LOG_ERR,"Writestream: write: %m");
			return(nwritten);	/* error */
		}

		nleft -= nwritten;
		ptr += nwritten;
	}
	return(nbytes - nleft);
}

/* from page 330, in UNIX Network Programming */
void Usleep(usec)
unsigned long usec;
{
	struct timeval	timeout;

	timeout.tv_sec  = usec / 1000000;
	timeout.tv_usec = usec - 1000000 * timeout.tv_sec;

	if (select(0, (fd_set *) 0, (fd_set *) 0, (fd_set *) 0, &timeout) < 0)
		logmsg(LOG_ERR,"Usleep: select: %m");
	return;
}

/*
 * The following routine was based on a Checksum function from CFDP.
 * The last four bytes have to be handled carefully.
 */

long Checksum(buf, len)
char *buf;				/* pointer to buffer */
u_int len;				/* sizeof buffer */
{
	register long sum=0;
	register int ilen;		/* how many ints in the buffer */
	register long *p;

	p = (long *)buf;
	ilen = len/(sizeof (long));

	while( ilen-- )
		sum += htonl(*p++);

	switch ( len % (sizeof (long)) ) {
	
	case 1:
		sum += htonl(*p) & 0xff000000;
		break;
	case 2:
		sum += htonl(*p) & 0xffff0000;
		break;
	case 3:
		sum += htonl(*p) & 0xffffff00;
		break;
	}

	return (-sum);
}

/*
 * The following ConfigureBroadcast() function sets up a linked list
 * of all the broadcast-capable interfaces on the local host.
 *
 * It originated from /usr/src/usr.sbin/rwhod/rwhod.c on the BSD Net2 tape.
 * I just removed what I did not need, and added some DEBUG calls.
 * - Steve Kotsopoulos <steve@ecf.toronto.edu>
 */

/*
 * Copyright (c) 1983 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* #ifndef lint
 * char copyright[] =
 * "@(#) Copyright (c) 1983 The Regents of the University of California.\n\
 All rights reserved.\n";
 * #endif
 *
 * #ifndef lint
 * static char sccsid[] = "@(#)rwhod.c	5.20 (Berkeley) 3/2/91";
 * #endif
 */

/*
 * Figure out device configuration and select
 * networks which deserve status information.
 */
int ConfigureBroadcast(s)
	int s;
{
	char buf[BUFSIZ], *cp, *cplim;
	struct ifconf ifc;
	struct ifreq ifreq, *ifr;
	struct sockaddr_in *sin;
	register struct neighbor *np;

	ifc.ifc_len = sizeof (buf);
	ifc.ifc_buf = buf;
	if (ioctl(s, SIOCGIFCONF, (char *)&ifc) < 0) {
		logmsg(LOG_ERR, "ioctl (get interface configuration)");
		return (0);
	}
	ifr = ifc.ifc_req;
#ifdef AF_LINK
#define max(a, b) (a > b ? a : b)
#define size(p)	max((p).sa_len, sizeof(p))
#else
#define size(p) (sizeof (p))
#endif
	cplim = buf + ifc.ifc_len; /*skip over if's with big ifr_addr's */
	for (cp = buf; cp < cplim;
			cp += sizeof (ifr->ifr_name) + size(ifr->ifr_addr)) {
		ifr = (struct ifreq *)cp;
		for (np = neighbors; np != NULL; np = np->n_next)
			if (np->n_name &&
			    strcmp(ifr->ifr_name, np->n_name) == 0)
				break;
		if (np != NULL)
			continue;
		ifreq = *ifr;
		np = (struct neighbor *)malloc(sizeof (*np));
		if (np == NULL)
			continue;
		np->n_name = (char *)malloc(strlen(ifr->ifr_name) + 1);
		if (np->n_name == NULL) {
			free((char *)np);
			continue;
		}
		strcpy(np->n_name, ifr->ifr_name);
		logmsg(LOG_DEBUG,"Configure found: %s\n", np->n_name);
		np->n_addrlen = sizeof (ifr->ifr_addr);
		np->n_addr = (char *)malloc(np->n_addrlen);
		if (np->n_addr == NULL) {
			free(np->n_name);
			free((char *)np);
			continue;
		}
		bcopy((char *)&ifr->ifr_addr, np->n_addr, np->n_addrlen);
		if (ioctl(s, SIOCGIFFLAGS, (char *)&ifreq) < 0) {
			syslog(LOG_ERR, "ioctl (get interface flags)");
			free((char *)np);
			continue;
		}
		if ((ifreq.ifr_flags & IFF_UP) == 0 ||
		    (ifreq.ifr_flags & (IFF_BROADCAST|IFF_POINTOPOINT)) == 0) {
			logmsg(LOG_DEBUG,"rejecting: %s\n", ifr->ifr_name);
			free((char *)np);
			continue;
		}
		np->n_flags = ifreq.ifr_flags;
		if (np->n_flags & IFF_POINTOPOINT) {
			if (ioctl(s, SIOCGIFDSTADDR, (char *)&ifreq) < 0) {
				syslog(LOG_ERR, "ioctl (get dstaddr)");
				free((char *)np);
				continue;
			}
			/* we assume addresses are all the same size */
			bcopy((char *)&ifreq.ifr_dstaddr,
			  np->n_addr, np->n_addrlen);
		}
		if (np->n_flags & IFF_BROADCAST) {
			if (ioctl(s, SIOCGIFBRDADDR, (char *)&ifreq) < 0) {
				syslog(LOG_ERR, "ioctl (get broadaddr)");
				free((char *)np);
				continue;
			}
			/* we assume addresses are all the same size */
			bcopy((char *)&ifreq.ifr_broadaddr,
			  np->n_addr, np->n_addrlen);
		}
		/* gag, wish we could get rid of Internet dependencies */
		sin = (struct sockaddr_in *)np->n_addr;
		np->n_next = neighbors;
		neighbors = np;
		logmsg(LOG_INFO,"Configure using: %s\n", np->n_name);
	}
	return (1);
}
