/*
 *	Copyright (c) 1994, 1995 by Jeremy Cooperstock and Steve Kotsopoulos,
 *	All Rights Reserved.
 */

/*
 * subscriber.c - AFDP subscriber functions
 *
 * static char rcsid[] = "$Id: subscriber.c,v 1.51 1998/07/02 19:37:36 steve Exp $";
 */

#include "config.h"
#include "sockutils.h"
#include "afdp.h"
#include "afdplib.h"
#include "afdpjoin.h"
#ifdef ACL
#include "acl.h"
#endif
#include <stdlib.h>			/* for srand(), rand() */
#include <unistd.h>

/* subscriber globals */
time_t join_time;			/* timestamp when I joined group */

/* subscriber message list stuff */
typedef struct msg_entry {
	msg_id_t msg_id;
	boolean_t New;
	boolean_t skip;
	host_t client_host;
	port_t client_port;
	char tmpfname[MAX_FILENAME];
	char filename[MAX_FILENAME];
	char cmdname[MAX_CMDNAME];
	char options;
	FILE *fp;
	int timeouts;
	int backoff_count;
	seq_t seq_total;
	seq_t packets_since_timeout;
	seq_t first_missing;
	boolean_t *received;	/* calloc'd array TRUE for each packet rcvd */
	struct msg_entry *prev;
	struct msg_entry *next;
} msg_entry_t;

msg_entry_t *msglist_head = NULL;

/* internal function prototypes */
msg_entry_t *WhichMsg();
void SequenceCheck(), Slowdown();
int RequestResend();
void FreeEntry();
boolean_t RandomBackoff();
void ResetBackoff();

int Join(group)
group_id_t group;
{
	int sock;
	admin_packet_t packet;

	packet.packet_type = JOIN_REQUEST;
	packet.group_id = htonl (group);
#ifdef MULTICAST
	packet.capability = MODE_MULTICAST;
#else
	packet.capability = MODE_BROADCAST;
#endif
	if (TCPexchange(secretary_host, &sock, admin_port, &packet, 
			sizeof(admin_packet_t)) != SUCCESS) {
		logmsg(LOG_INFO,"Join: FAIL from xchange admin packet\n");
		return( FAIL );
	}
	close(sock);

	/* process reply packet */
	if (packet.packet_type != JOIN_REPLY) {
		logmsg(LOG_ERR,
			"Join expected packet type %d, received %d\n",
			JOIN_REPLY, packet.packet_type);
		return( FAIL );
	}

	if (packet.status != SUCCESS) {
		logmsg(LOG_ERR, "RequestJoin failed\n");
		return( FAIL );
	}

	join_time = ntohl(packet.time);
	group_data_port = ntohs(packet.group_data_port);
	return (SUCCESS);
}

int Leave()
{
	int sock;
	admin_packet_t packet;

	logmsg(LOG_DEBUG,"Leave:  entering\n");
	packet.packet_type = LEAVE_REQUEST;
	packet.group_id = htonl(mygroup);
	if (TCPexchange(secretary_host, &sock, admin_port, &packet, 
			sizeof(admin_packet_t)) != SUCCESS) {
		logmsg(LOG_INFO,"Leave: returned FAIL from sending admin packet\n");
		return( FAIL );
	}
	close(sock);

	if (packet.packet_type != LEAVE_REPLY) {
		logmsg(LOG_ERR, "Leave: expected packet type %d, received %d\n",
			LEAVE_REPLY, packet.packet_type);
		return( FAIL );
	}
	return(SUCCESS);
}

void HandleGroupDataPacket()
{
	int i, got, len;
	data_packet_t packet;
	struct sockaddr_in from;
	seq_t seq_number;
	data_t *data;
	u_short data_size;
	msg_entry_t *entry;
	long ch;

	len = sizeof (from);
	got = recvfrom(group_data_sock, (char *)&packet, 
		sizeof (data_packet_t), 0, (struct sockaddr *)&from, &len);

	logmsg(LOG_DEBUG,"HandleGroupDataPacket: got packet from %s (port %u)\n",
		HostaddrToName(from.sin_addr.s_addr), ntohs(from.sin_port));

#ifdef ACL
	if (!CheckAuth(&packet, got, &from))
		return;
#endif

	if (packet.packet_type == SHUTDOWN) {
		msg_entry_t *p;
		control_packet_t shutdown;

		logmsg(LOG_NOTICE, "Got SHUTDOWN message from %s\n",
				HostaddrToName(from.sin_addr.s_addr));
		if (from.sin_addr.s_addr != secretary_host) {
			logmsg(LOG_INFO,"ignoring SHUTDOWN, secretary is %s\n",
			HostaddrToName(secretary_host));
			return;
		}
			
		/* now, send a SHUTDOWN packet to each active
		 * client (publisher), before we exit
		 */
		shutdown.packet_type = SHUTDOWN;

		p = msglist_head;
		while (p != NULL) {
			TCPsend(p->client_host, p->client_port, 
				(char *)&shutdown, sizeof(control_packet_t));
			p = p->next;
		}
		exit(0);
	}

	/* verify that we got right number of bytes */
	if (ntohs(packet.data_size) + DATA_PACKET_NONDATA_SIZE != got) {
		logmsg(LOG_NOTICE,"received %d bytes, expecting %d\n", got, 
		  (int) (ntohs(packet.data_size) + DATA_PACKET_NONDATA_SIZE));
	}

	/* verify packet type */
	if (packet.packet_type != MSG_SEND && 
	    packet.packet_type != MSG_RESEND &&
	    packet.packet_type != MSG_SKIP) {
		logmsg(LOG_NOTICE,"message packet of wrong type: %d\n",
			packet.packet_type);
		return;
	}

	/* verify group */
	if (ntohl (packet.group_id) != mygroup) {
		logmsg(LOG_NOTICE,"%s %lu, expecting %lu\n",
			"message packet for wrong group: ",
			ntohl(packet.group_id), mygroup);
		return;
	}

	/* verify publisher, if we must */
	if (publisher_host && (from.sin_addr.s_addr != publisher_host)) {
		logmsg(LOG_NOTICE,"publisher host %s rejected\n",
			HostaddrToName(from.sin_addr.s_addr));
		return;
	}

	/* verify that we should still be receiving this message */
	entry = WhichMsg(ntohs(packet.msg_id));
	if (entry->skip) {
		logmsg(LOG_INFO,"received packet for msg being skipped (msg %d)\n", 
			 ntohs(packet.msg_id));
		return;
	}
	
	/*
	 * If publisher will not resend, skip remainder of message.
	 * We need to remove the temporary file and also cancel the
	 * timer which calls SequenceCheck.
	 */
	if (packet.packet_type == MSG_SKIP) {
		logmsg(LOG_NOTICE,"publisher refuses to resend. Skipping msg %d\n",
			 ntohs(packet.msg_id));
		fclose(entry->fp);
		unlink(entry->tmpfname);
		entry->skip = TRUE;
		CancelTimer (entry->msg_id);
	}

	/* verify publish time */
	if (ntohl (packet.time) <= join_time) {
		logmsg(LOG_NOTICE,"packet timestamp earlier than join time");
		return;
	}

	/*
	 * If this is the first packet we have seen of the current
	 * message, then initialize the appropriate entries.
	 * Note that we use calloc() to dynamically allocate space for
	 * the received list and ensure that it will be initialized to NULL. 
	 * Also, note that we must allocate space for seq_total + 1 entries
	 * since the sequence numbers begin at 1!
	 * Then, assign a temporary filename and open for output.
	 * Finally, start the timer to occasionally call SequenceCheck.
	 */

	i = seq_number = ntohl(packet.seq_number);

	if (entry->New) {
		logmsg(LOG_INFO,"Received packet from a new message.\n");
		logmsg(LOG_INFO,"seq_number is %lu, seq_total is %lu, msg is %u\n", 
		    seq_number, ntohl(packet.seq_total), ntohs(packet.msg_id));
		entry->New = FALSE;
		entry->client_host = from.sin_addr.s_addr;
		entry->client_port = packet.control_port;
		entry->options = packet.options;
		entry->timeouts = 0;
		ResetBackoff(entry);
		entry->seq_total = ntohl(packet.seq_total);	
		entry->first_missing = 1;  /* don't know if we have 1 yet! */
		entry->packets_since_timeout = 0;
		entry->received = (boolean_t *)
			calloc(entry->seq_total+1, sizeof(boolean_t));
	}

	/* verify sequence total */
	else if (ntohl(packet.seq_total) != entry->seq_total) 
		logmsg(LOG_NOTICE,"Sequence total altered - proceeding anyway\n"); 

	/* verify sequence number */
	if (seq_number > entry->seq_total) {
		logmsg(LOG_NOTICE,"Sequence number %lu out of range - ignoring\n",
			seq_number);
		return;
	}

	/* check for duplicates */
	if (entry->received[seq_number] == TRUE) {
		logmsg(LOG_INFO,"Duplicate packet...ignoring\n");
		return;
	}

	/* verify checksum */
	ch = Checksum (packet.data, ntohs(packet.data_size));
	if (ch != ntohl(packet.checksum)) {
		logmsg(LOG_NOTICE,"Checksum error: expected %ld got %ld\n", 
			ch, ntohl(packet.checksum));
		return; 
	}

	/* check for packet 1 */
	if (seq_number == 1) {
		strncpy (entry->filename, (char *)packet.data, MAX_FILENAME);
		strncpy (entry->cmdname, (char *)&(packet.data[MAX_FILENAME]), 
			MAX_CMDNAME);
		if(safe_command) {
			logmsg(LOG_NOTICE,"forcing cmdname to %s",safe_command);
			strncpy (entry->cmdname, safe_command, MAX_CMDNAME);
		}

		sprintf (entry->tmpfname, "%s.%s.%u", entry->filename,
			"tmp", ntohs(packet.msg_id));
		entry->fp = efopen(entry->tmpfname, "w");

		data = &(packet.data[sizeof(file_info_t)]);
		data_size = ntohs(packet.data_size) - sizeof(file_info_t);
		StartTimer (entry->msg_id, WAIT_TIME, SequenceCheck, entry);
	}

	/* normal packet */
	else { 
		if(!entry->fp) {
			/* we must have dropped the first packet */
			logmsg(LOG_ERR,"dropped first packet, RequestResend");
			RequestResend(entry, 1, 1);
			return;
		}
		data = packet.data;
		data_size = ntohs(packet.data_size);
	}

	/* seek to correct position and write data */
	logmsg(LOG_INFO,"writing packet %lu:  %u bytes at offset %lu\n", 
			seq_number, data_size, ntohl(packet.offset));

	fseek(entry->fp, ntohl(packet.offset), 0);
	fwrite(data, data_size, 1, entry->fp);

	/* mark packet received */
	entry->received[seq_number] = TRUE;
	if (packet.packet_type != MSG_RESEND) {
	    while ((--i >= 0) && (i > (seq_number - 10))) {
		if(entry->received[i] == FALSE) {
			Slowdown(entry);
			break;
		}
	    }
	}

	/* update timeout variables */
	entry->timeouts = 0;
	entry->packets_since_timeout++;
}

/*
 * Subscriber utility functions 
 */


/*
 * Slowdown is called whenever we detect we've missed a packet
 */
void Slowdown(entry)
msg_entry_t *entry;
{
	control_packet_t packet;

	logmsg(LOG_DEBUG,"Slowdown called for msg %u\n", entry->msg_id);

	packet.packet_type = SLOWDOWN_REQUEST;
	packet.msg_id = htons(entry->msg_id);

	if (TCPsend(entry->client_host, entry->client_port, 
			(char *)&packet, sizeof(control_packet_t)) == -1) {
		logmsg(LOG_NOTICE,"can't contact publisher. Skipping msg %d\n",
			ntohs(packet.msg_id));
		fclose(entry->fp);
		unlink(entry->tmpfname);
		entry->skip = TRUE;
		CancelTimer (entry->msg_id);
		return;
	}
}

/*
 * SequenceCheck is called periodically for each current
 * message, regardless of the number of packet events since the last call.
 */
void SequenceCheck(entry)
msg_entry_t *entry;
{
	int i;
 	int low = 0;
	int high = 0;
	char cmd[MAX_CMDNAME + MAX_FILENAME + 1];

	logmsg(LOG_DEBUG,"sequence check called for msg %u\n", entry->msg_id);

	for (i = entry->first_missing; i <= entry->seq_total; i++) {
		if (entry->received[i] == FALSE) {
			entry->first_missing = low = high = i;
			break;
		}
	}

	if (low) {
		entry->timeouts++;	/* we've been waiting one more unit */

		do 
			high++;
		while ( high <= entry->seq_total && 
			high <= low + MAX_RESEND_PACKETS &&
			entry->received[high] == FALSE);


		/*
		 * If either some packets are currently out of order
		 * or no packets have arrived since the last timeout
		 * and we are still waiting for the remaining packets 
		 * of the message, then issue a resend request, clear
		 * the timeouts counter and restart timer.
		 */

		if (entry->received[high] == TRUE ||
	    		   (entry->packets_since_timeout == 0 && 
	      		    entry->timeouts > MAX_TIMEOUTS)) {

			/*
		 	 * We only worry about generating a resend request
		 	 * if the RandomBackoff() function doesn't object.
			 * Otherwise, count down and reset the timer.
		 	 */

			if (RandomBackoff(entry) == FALSE) {
				high--;

				if (RequestResend (entry, low, high) != SUCCESS)
					return;

				entry->timeouts = 0;
				StartTimer (entry->msg_id, RESEND_REQUEST_TIME, 
						SequenceCheck, entry);
			}

			else {
				StartTimer (entry->msg_id, WAIT_TIME, 
						SequenceCheck, entry);
			}
		}
				

		else {
			ResetBackoff(entry);
			StartTimer (entry->msg_id, WAIT_TIME, 
					SequenceCheck, entry);
		}

		/* now reset count of packets received since last timeout */
		entry->packets_since_timeout = 0;
	}
					
	else {
		logmsg(LOG_DEBUG,"file received: close and prepare to free entry\n");
		fclose(entry->fp);
		entry->skip = TRUE; /* ignore any further pkts from this msg */
		if (rename (entry->tmpfname, entry->filename) != 0)
			fatal("SequenceCheck: rename failed: %m");
		
		if (entry->cmdname[0] != (char) NULL) {
			sprintf (cmd, "%s %s", entry->cmdname, entry->filename);
			logmsg(LOG_INFO,"about to run command: %s\n", cmd);
			system (cmd);
		}

		else
			logmsg(LOG_INFO,"no command to run... just saving file, %s\n",
				entry->filename);

		if (entry->options && OPTION_CONFIRM_RECEIPT)
			logmsg(LOG_NOTICE,"Options: should CONFIRM RECEIPT");

		/* now set timer to remove entry in message table */
		StartTimer (entry->msg_id, FREE_TIME, FreeEntry, entry);
	}
}

int RequestResend(entry, low, high)
msg_entry_t *entry;
int low, high;
{
	control_packet_t packet;

	packet.packet_type = RESEND_REQUEST;
	packet.msg_id = htons(entry->msg_id);
	packet.seq_number_low = htonl(low);
	packet.seq_number_high = htonl(high);

	logmsg(LOG_NOTICE,"ask for resend of packets %d - %d, msg %d\n", 
		low, high, entry->msg_id);

	if (TCPsend(entry->client_host, entry->client_port, 
	       (char *)&packet, sizeof(control_packet_t)) == -1) {
		logmsg(LOG_NOTICE,"can't contact publisher. Skipping msg %d\n",
		ntohs(packet.msg_id));
		fclose(entry->fp);
		unlink(entry->tmpfname);
		entry->skip = TRUE;
		return (FAIL);
	}

	else
		return (SUCCESS);
}

void FreeEntry(entry)
msg_entry_t *entry;
{
	logmsg(LOG_DEBUG,"now freeing entry for msg %u\n", entry->msg_id);
	if (entry->next)
		entry->next->prev = entry->prev;
	if (entry->prev)
		entry->prev->next = entry->next;	
	free(entry->received);

	if (entry == msglist_head)
		msglist_head = entry->next;
	free(entry);
}

msg_entry_t *WhichMsg(msg_id)
msg_id_t msg_id;
{
	msg_entry_t *p; 

	logmsg(LOG_DEBUG,"WhichMsg called for msg_id %u\n", msg_id);
	if (msglist_head) {
		logmsg(LOG_DEBUG,"head currently msg %u\n", msglist_head->msg_id);

		p = msglist_head;
		while (p->msg_id != msg_id) {
			if (!p->next) {
				p->next = (msg_entry_t *) malloc (
					sizeof(msg_entry_t));
				p->next->msg_id = msg_id;
				p->next->New = TRUE;
				p->next->skip = FALSE;
				p->next->prev = p;
				p->next->next = NULL;
				p->next->fp = NULL;
			}
			else
				p = p->next;
		}
	}

	else {
		logmsg(LOG_DEBUG,"head currently NULL\n");

		p = msglist_head = (msg_entry_t *) malloc (sizeof(msg_entry_t));
		if (!p)
			fatal("Could not malloc entry for message: %m");
		p->msg_id = msg_id;
		p->New = TRUE;
		p->skip = FALSE;
		p->prev = NULL;
		p->next = NULL;
		p->fp = NULL;
	}
	return(p);
}

/*
 * quick-and-dirty random wait routine -
 * We don't want to queue up a resend request for a later time, 
 * since the publisher might deliver in the interim.  Instead, 
 * generate a random number mod MAX_BACKOFF_COUNT and count this 
 * down until 0.  In the meantime, pretend all is well.
 * 
 * It might make sense to modify this number based on the
 * count of active subscribers.  Right now, we don't have
 * this information available.  Another improvement might 
 * be to use a non-linear random distribution (exponential?).
 */

boolean_t RandomBackoff(entry) 
msg_entry_t *entry;
{
	static mypid = 0;
	boolean_t backoff = TRUE;

	if (opt_random_wait == TRUE) {
		if (!mypid) {
			mypid = getpid();
			srand(mypid);
		}

		if (entry->backoff_count < 0) 
			entry->backoff_count = rand() % MAX_BACKOFF_COUNT;

		else if (entry->backoff_count > 0)
			entry->backoff_count--;

		else
			backoff = FALSE;
	}
	return (backoff);
}

void ResetBackoff(entry) 
msg_entry_t *entry;
{
	entry->backoff_count = -1;
}
